from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any, Generic, TypeVar

import hydra
import torch
from hydra.core.hydra_config import HydraConfig
from lightning_fabric.utilities.device_parser import _parse_gpu_ids
from omegaconf import DictConfig, OmegaConf, open_dict
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.loggers import Logger, MLFlowLogger
from pytorch_lightning.utilities.exceptions import MisconfigurationException

from quadra import get_version
from quadra.callbacks.mlflow import validate_artifact_storage
from quadra.datamodules.base import BaseDataModule
from quadra.models.evaluation import BaseEvaluationModel
from quadra.utils import utils
from quadra.utils.export import import_deployment_model

log = utils.get_logger(__name__)
DataModuleT = TypeVar("DataModuleT", bound=BaseDataModule)


class Task(Generic[DataModuleT]):
    """Base Experiment Task.

    Args:
        config: The experiment configuration.
    """

    def __init__(self, config: DictConfig):
        self.config = config
        self.export_folder: str = "deployment_model"
        self._datamodule: DataModuleT
        self.metadata: dict[str, Any]
        self.save_config()

    def save_config(self) -> None:
        """Save the experiment configuration when running an Hydra experiment."""
        if HydraConfig.initialized():
            with open("config_resolved.yaml", "w") as fp:
                OmegaConf.save(config=OmegaConf.to_container(self.config, resolve=True), f=fp.name)

    def prepare(self) -> None:
        """Prepare the experiment."""
        self.datamodule = self.config.datamodule

    @property
    def datamodule(self) -> DataModuleT:
        """T_DATAMODULE: The datamodule."""
        return self._datamodule

    @datamodule.setter
    def datamodule(self, datamodule_config: DictConfig) -> None:
        """DataModuleT: The datamodule. Instantiated from the datamodule config."""
        log.info("Instantiating datamodule <%s>", {datamodule_config["_target_"]})
        datamodule: DataModuleT = hydra.utils.instantiate(datamodule_config)
        self._datamodule = datamodule

    def train(self) -> Any:
        """Train the model."""
        log.info("Training not implemented for this task!")

    def test(self) -> Any:
        """Test the model."""
        log.info("Testing not implemented for this task!")

    def export(self) -> None:
        """Export model for production."""
        log.info("Export model for production not implemented for this task!")

    def generate_report(self) -> None:
        """Generate a report."""
        log.info("Report generation not implemented for this task!")

    def finalize(self) -> None:
        """Finalize the experiment."""
        log.info("Results are saved in %s", os.getcwd())

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.train()
        self.test()
        if self.config.export is not None and len(self.config.export.types) > 0:
            self.export()
        self.generate_report()
        self.finalize()


class LightningTask(Generic[DataModuleT], Task[DataModuleT]):
    """Base Experiment Task.

    Args:
        config: The experiment configuration
        checkpoint_path: The path to the checkpoint to load the model from. Defaults to None.
        run_test: Whether to run the test after training. Defaults to False.
        report: Whether to generate a report. Defaults to False.
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
        report: bool = False,
    ):
        super().__init__(config=config)
        self.checkpoint_path = checkpoint_path
        self.run_test = run_test
        self.report = report
        self._module: LightningModule
        self._devices: int | list[int]
        self._callbacks: list[Callback]
        self._logger: list[Logger]
        self._trainer: Trainer

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()

        # First setup loggers since some callbacks might need logger setup correctly.
        if "logger" in self.config:
            self.logger = self.config.logger

        if "callbacks" in self.config:
            self.callbacks = self.config.callbacks

        self.devices = self.config.trainer.devices
        self.trainer = self.config.trainer

    @property
    def module(self) -> LightningModule:
        """LightningModule: The model."""
        return self._module

    @module.setter
    def module(self, module_config) -> None:
        """LightningModule: The model."""
        raise NotImplementedError("module must be set in subclass")

    @property
    def trainer(self) -> Trainer:
        """Trainer: The trainer."""
        return self._trainer

    @trainer.setter
    def trainer(self, trainer_config: DictConfig) -> None:
        """Trainer: The trainer."""
        log.info("Instantiating trainer <%s>", trainer_config["_target_"])
        trainer_config.devices = self.devices
        trainer: Trainer = hydra.utils.instantiate(
            trainer_config,
            callbacks=self.callbacks,
            logger=self.logger,
            _convert_="partial",
        )
        self._trainer = trainer

    @property
    def callbacks(self) -> list[Callback]:
        """List[Callback]: The callbacks."""
        return self._callbacks

    @callbacks.setter
    def callbacks(self, callbacks_config) -> None:
        """List[Callback]: The callbacks."""
        if self.config.core.get("unit_test"):
            log.info("Unit Testing, skipping callbacks")
            return
        instatiated_callbacks = []
        for _, cb_conf in callbacks_config.items():
            if "_target_" in cb_conf:
                # Disable is a reserved keyword for callbacks, hopefully no callback will use it
                if "disable" in cb_conf:
                    if cb_conf["disable"]:
                        log.info("Skipping callback <%s> as it is disabled", cb_conf["_target_"])
                        continue

                    with open_dict(cb_conf):
                        del cb_conf.disable

                # Skip the gpu stats logger callback if no gpu is available to avoid errors
                if not torch.cuda.is_available() and cb_conf["_target_"] == "nvitop.callbacks.lightning.GpuStatsLogger":
                    continue

                log.info("Instantiating callback <%s>", cb_conf["_target_"])
                instatiated_callbacks.append(hydra.utils.instantiate(cb_conf))
        self._callbacks = instatiated_callbacks
        if len(instatiated_callbacks) <= 0:
            log.warning("No callback found in configuration.")

    @property
    def logger(self) -> list[Logger]:
        """List[Logger]: The loggers."""
        return self._logger

    @logger.setter
    def logger(self, logger_config) -> None:
        """List[Logger]: The loggers."""
        if self.config.core.get("unit_test"):
            log.info("Unit Testing, skipping loggers")
            return
        instantiated_loggers = []
        for _, lg_conf in logger_config.items():
            if "_target_" in lg_conf:
                log.info("Instantiating logger <%s>", lg_conf["_target_"])
                logger = hydra.utils.instantiate(lg_conf)
                if isinstance(logger, MLFlowLogger):
                    validate_artifact_storage(logger)
                instantiated_loggers.append(logger)

        self._logger = instantiated_loggers

        if len(instantiated_loggers) <= 0:
            log.warning("No logger found in configuration.")

    @property
    def devices(self) -> int | list[int]:
        """List[int]: The devices ids."""
        return self._devices

    @devices.setter
    def devices(self, devices) -> None:
        """List[int]: The devices ids."""
        if self.config.trainer.get("accelerator") == "cpu":
            self._devices = self.config.trainer.devices
            return

        try:
            self._devices = _parse_gpu_ids(devices, include_cuda=True)
        except MisconfigurationException:
            self._devices = 1
            self.config.trainer["accelerator"] = "cpu"
            log.warning("Trying to instantiate GPUs but no GPUs are available, training will be done on CPU")

    def train(self) -> None:
        """Train the model."""
        log.info("Starting training!")
        utils.log_hyperparameters(
            config=self.config,
            model=self.module,
            trainer=self.trainer,
        )

        self.trainer.fit(model=self.module, datamodule=self.datamodule)

    def test(self) -> Any:
        """Test the model."""
        log.info("Starting testing!")

        best_model = None
        if (
            self.trainer.checkpoint_callback is not None
            and hasattr(self.trainer.checkpoint_callback, "best_model_path")
            and self.trainer.checkpoint_callback.best_model_path is not None
            and len(self.trainer.checkpoint_callback.best_model_path) > 0
        ):
            best_model = self.trainer.checkpoint_callback.best_model_path

        if best_model is None:
            log.warning(
                "No best checkpoint model found, using last weights for test, this might lead to worse results, "
                "consider using a checkpoint callback."
            )

        return self.trainer.test(model=self.module, datamodule=self.datamodule, ckpt_path=best_model)

    def finalize(self) -> None:
        """Finalize the experiment."""
        super().finalize()
        utils.finish(
            config=self.config,
            module=self.module,
            datamodule=self.datamodule,
            trainer=self.trainer,
            callbacks=self.callbacks,
            logger=self.logger,
            export_folder=self.export_folder,
        )

        if (
            not self.config.trainer.get("fast_dev_run")
            and self.trainer.checkpoint_callback is not None
            and hasattr(self.trainer.checkpoint_callback, "best_model_path")
        ):
            log.info("Best model ckpt: %s", self.trainer.checkpoint_callback.best_model_path)

    def add_callback(self, callback: Callback):
        """Add a callback to the trainer.

        Args:
            callback: The callback to add
        """
        if hasattr(self.trainer, "callbacks") and isinstance(self.trainer.callbacks, list):
            self.trainer.callbacks.append(callback)

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.train()
        if self.run_test:
            self.test()
        if self.config.export is not None and len(self.config.export.types) > 0:
            self.export()
        if self.report:
            self.generate_report()
        self.finalize()


class PlaceholderTask(Task):
    """Placeholder task."""

    def execute(self) -> None:
        """Execute the task and all the steps."""
        log.info("Running Placeholder Task.")
        log.info("Quadra Version: %s", str(get_version()))
        log.info("If you are reading this, it means that library is installed correctly!")


class Evaluation(Generic[DataModuleT], Task[DataModuleT]):
    """Base Evaluation Task with deployment models.

    Args:
        config: The experiment configuration
        model_path: The model path.
        device: Device to use for evaluation. If None, the device is automatically determined.

    """

    def __init__(
        self,
        config: DictConfig,
        model_path: str,
        device: str | None = None,
    ):
        super().__init__(config=config)

        if device is None:
            self.device = utils.get_device()
        else:
            self.device = device

        self.config = config
        self.model_data: dict[str, Any]
        self.model_path = model_path
        self._deployment_model: BaseEvaluationModel
        self.deployment_model_type: str
        self.model_info_filename = "model.json"
        self.report_path = ""
        self.metadata = {"report_files": []}

    @property
    def deployment_model(self) -> BaseEvaluationModel:
        """Deployment model."""
        return self._deployment_model

    @deployment_model.setter
    def deployment_model(self, model_path: str):
        """Set the deployment model."""
        self._deployment_model = import_deployment_model(
            model_path=model_path, device=self.device, inference_config=self.config.inference
        )

    def prepare(self) -> None:
        """Prepare the evaluation."""
        with open(os.path.join(Path(self.model_path).parent, self.model_info_filename)) as f:
            self.model_data = json.load(f)

        if not isinstance(self.model_data, dict):
            raise ValueError("Model info file is not a valid json")

        for input_size in self.model_data["input_size"]:
            if len(input_size) != 3:
                continue

            # Adjust the transform for 2D models (CxHxW)
            # We assume that each input size has the same height and width
            if input_size[1] != self.config.transforms.input_height:
                log.warning(
                    "Input height of the model (%s) is different from the one specified "
                    + "in the config (%s). Fixing the config.",
                    input_size[1],
                    self.config.transforms.input_height,
                )
                self.config.transforms.input_height = input_size[1]

            if input_size[2] != self.config.transforms.input_width:
                log.warning(
                    "Input width of the model (%s) is different from the one specified "
                    + "in the config (%s). Fixing the config.",
                    input_size[2],
                    self.config.transforms.input_width,
                )
                self.config.transforms.input_width = input_size[2]

        self.deployment_model = self.model_path  # type: ignore[assignment]



# pylint: disable=unsubscriptable-object,unsupported-assignment-operation,unsupported-membership-test
from __future__ import annotations

import glob
import os
import random
from typing import Any

import albumentations
import cv2
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from skmultilearn.model_selection import iterative_train_test_split
from torch.utils.data import DataLoader

from quadra.datamodules.base import BaseDataModule
from quadra.datasets.segmentation import SegmentationDataset, SegmentationDatasetMulticlass
from quadra.utils import utils

log = utils.get_logger(__name__)


class SegmentationDataModule(BaseDataModule):
    """Base class for segmentation datasets.

    Args:
        data_path: Path to the data main folder.
        name: The name for the data module. Defaults to "segmentation_datamodule".
        val_size: The validation split. Defaults to 0.2.
        test_size: The test split. Defaults to 0.2.
        seed: Random generator seed. Defaults to 42.
        dataset: Dataset class.
        batch_size: Batch size. Defaults to 32.
        num_workers: Number of workers for dataloaders. Defaults to 16.
        train_transform: Transformations for train dataset.
            Defaults to None.
        val_transform: Transformations for validation dataset.
            Defaults to None.
        test_transform: Transformations for test dataset.
            Defaults to None.
        num_data_class: The number of samples per class. Defaults to None.
        exclude_good: If True, exclude good samples from the dataset. Defaults to False.
    """

    def __init__(
        self,
        data_path: str,
        name: str = "segmentation_datamodule",
        test_size: float = 0.3,
        val_size: float = 0.3,
        seed: int = 42,
        dataset: type[SegmentationDataset] = SegmentationDataset,
        batch_size: int = 32,
        num_workers: int = 6,
        train_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        train_split_file: str | None = None,
        test_split_file: str | None = None,
        val_split_file: str | None = None,
        num_data_class: int | None = None,
        exclude_good: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            batch_size=batch_size,
            num_workers=num_workers,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            **kwargs,
        )
        self.test_size = test_size
        self.val_size = val_size
        self.num_data_class = num_data_class
        self.exclude_good = exclude_good
        self.train_split_file = train_split_file
        self.test_split_file = test_split_file
        self.val_split_file = val_split_file
        self.dataset = dataset
        self.train_dataset: SegmentationDataset
        self.val_dataset: SegmentationDataset
        self.test_dataset: SegmentationDataset

    def _preprocess_mask(self, mask) -> np.ndarray:
        """Binarize mask using 0 as threshold."""
        mask = (mask > 0).astype(np.uint8)
        return mask

    @staticmethod
    def _resolve_label(path: str) -> int:
        """Resolve label from mask.

        Args:
            path: Path to the mask.

        Returns:
            0 if the mask is empty, 1 otherwise
        """
        if cv2.imread(path).sum() == 0:
            return 0

        return 1

    def _read_folder(self, data_path: str) -> tuple[list[str], list[int], list[str]]:
        """Read a folder containing images and masks subfolders.

        Args:
            data_path: Path to the data folder.

        Returns:
            List of paths to the images, associated binary targets and list to paths to the masks.
        """
        samples = []
        targets = []
        masks = []

        for im in glob.glob(os.path.join(data_path, "images", "*")):
            if im[0] == ".":
                continue

            mask_path = glob.glob(os.path.splitext(im.replace("images", "masks"))[0] + ".*")

            if len(mask_path) == 0:
                log.debug("Mask not found: %s", os.path.basename(im))
                continue

            if len(mask_path) > 1:
                raise ValueError(f"Multiple masks found for image: {os.path.basename(im)}, this is not supported")

            target = self._resolve_label(mask_path[0])
            samples.append(im)
            targets.append(target)
            masks.append(mask_path[0])

        return samples, targets, masks

    def _read_split(self, split_file: str) -> tuple[list[str], list[int], list[str]]:
        """Reads split file.

        Args:
            split_file: Path to the split file.

        Returns:
            List of paths to images, List of labels.
        """
        samples, targets, masks = [], [], []
        with open(split_file) as f:
            split = f.read().splitlines()
        for sample in split:
            sample_path = os.path.join(self.data_path, sample)
            mask_path = glob.glob(os.path.splitext(sample_path.replace("images", "masks"))[0] + ".*")

            if len(mask_path) == 0:
                log.debug("Mask not found: %s", os.path.basename(sample_path))
                continue

            if len(mask_path) > 1:
                raise ValueError(
                    f"Multiple masks found for image: {os.path.basename(sample_path)}, this is not supported"
                )

            target = self._resolve_label(mask_path[0])
            samples.append(sample_path)
            targets.append(target)
            masks.append(mask_path[0])

        return samples, targets, masks

    def _prepare_data(self) -> None:
        """Prepare data for training and testing."""
        if not (self.test_split_file and self.train_split_file and self.val_split_file):
            all_samples, all_targets, all_masks = self._read_folder(self.data_path)
            samples_train, samples_test, targets_train, targets_test, masks_train, masks_test = train_test_split(
                all_samples,
                all_targets,
                all_masks,
                test_size=self.test_size,
                random_state=self.seed,
                stratify=all_targets,
            )
        if self.test_split_file:
            samples_test, targets_test, masks_test = self._read_split(self.test_split_file)
            if not self.train_split_file:
                samples_train, targets_train, masks_train = [], [], []
                for sample, target, mask in zip(all_samples, all_targets, all_masks, strict=False):
                    if sample not in samples_test:
                        samples_train.append(sample)
                        targets_train.append(target)
                        masks_train.append(mask)

        if self.train_split_file:
            samples_train, targets_train, masks_train = self._read_split(self.train_split_file)
            if not self.test_split_file:
                samples_test, targets_test, masks_test = [], [], []
                for sample, target, mask in zip(all_samples, all_targets, all_masks, strict=False):
                    if sample not in samples_train:
                        samples_test.append(sample)
                        targets_test.append(target)
                        masks_test.append(mask)

        if self.val_split_file:
            if not self.test_split_file or not self.train_split_file:
                raise ValueError("Validation split file is specified but no train or test split file is specified.")
            samples_val, targets_val, masks_val = self._read_split(self.val_split_file)
        else:
            samples_train, samples_val, targets_train, targets_val, masks_train, masks_val = train_test_split(
                samples_train,
                targets_train,
                masks_train,
                test_size=self.val_size,
                random_state=self.seed,
                stratify=targets_train,
            )

        if self.exclude_good:
            samples_train = list(np.array(samples_train)[np.array(targets_train) != 0])
            masks_train = list(np.array(masks_train)[np.array(targets_train) != 0])
            targets_train = list(np.array(targets_train)[np.array(targets_train) != 0])

        if self.num_data_class is not None:
            samples_train_topick = []
            targets_train_topick = []
            masks_train_topick = []

            for cl in np.unique(targets_train):
                idx = np.where(np.array(targets_train) == cl)[0].tolist()
                random.seed(self.seed)
                random.shuffle(idx)
                to_pick = idx[: self.num_data_class]
                for i in to_pick:
                    samples_train_topick.append(samples_train[i])
                    targets_train_topick.append(cl)
                    masks_train_topick.append(masks_train[i])

            samples_train = samples_train_topick
            targets_train = targets_train_topick
            masks_train = masks_train_topick

        df_list = []
        for split_name, samples, targets, masks in [
            ("train", samples_train, targets_train, masks_train),
            ("val", samples_val, targets_val, masks_val),
            ("test", samples_test, targets_test, masks_test),
        ]:
            df = pd.DataFrame({"samples": samples, "targets": targets, "masks": masks})
            df["split"] = split_name
            df_list.append(df)

        self.data = pd.concat(df_list, axis=0)

    def setup(self, stage=None):
        """Setup data module based on stages of training."""
        if stage in ["fit", "train"]:
            self.train_dataset = self.dataset(
                image_paths=self.data[self.data["split"] == "train"]["samples"].tolist(),
                mask_paths=self.data[self.data["split"] == "train"]["masks"].tolist(),
                mask_preprocess=self._preprocess_mask,
                labels=self.data[self.data["split"] == "train"]["targets"].tolist(),
                object_masks=None,
                transform=self.train_transform,
                batch_size=None,
                defect_transform=None,
                resize=None,
            )
            self.val_dataset = self.dataset(
                image_paths=self.data[self.data["split"] == "val"]["samples"].tolist(),
                mask_paths=self.data[self.data["split"] == "val"]["masks"].tolist(),
                defect_transform=None,
                labels=self.data[self.data["split"] == "val"]["targets"].tolist(),
                object_masks=None,
                batch_size=None,
                mask_preprocess=self._preprocess_mask,
                transform=self.test_transform,
                resize=None,
            )
        elif stage == "test":
            self.test_dataset = self.dataset(
                image_paths=self.data[self.data["split"] == "test"]["samples"].tolist(),
                mask_paths=self.data[self.data["split"] == "test"]["masks"].tolist(),
                labels=self.data[self.data["split"] == "test"]["targets"].tolist(),
                object_masks=None,
                batch_size=None,
                mask_preprocess=self._preprocess_mask,
                transform=self.test_transform,
                resize=None,
            )
        elif stage == "predict":
            pass
        else:
            raise ValueError(f"Unknown stage {stage}")

    def train_dataloader(self) -> DataLoader:
        """Returns the train dataloader.

        Raises:
            ValueError: If train dataset is not initialized.

        Returns:
            Train dataloader.
        """
        if not self.train_dataset_available:
            raise ValueError("Train dataset is not initialized")

        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def val_dataloader(self) -> DataLoader:
        """Returns the validation dataloader.

        Raises:
            ValueError: If validation dataset is not initialized.

        Returns:
            val dataloader.
        """
        if not self.val_dataset_available:
            raise ValueError("Validation dataset is not initialized")

        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def test_dataloader(self) -> DataLoader:
        """Returns the test dataloader.

        Raises:
            ValueError: If test dataset is not initialized.


        Returns:
            test dataloader.
        """
        if not self.test_dataset_available:
            raise ValueError("Test dataset is not initialized")

        loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return self.test_dataloader()


class SegmentationMulticlassDataModule(BaseDataModule):
    """Base class for segmentation datasets with multiple classes.

    Args:
        data_path : Path to the data main folder.
        idx_to_class: dict with corrispondence btw mask index and classes: {1: class_1, 2: class_2, ..., N: class_N}
            except background class which is 0.
        name : The name for the data module. Defaults to "multiclass_segmentation_datamodule".
        dataset: Dataset class.
        batch_size : Batch size. Defaults to 32.
        val_size : The validation split. Defaults to 0.3.
        test_size : The test split. Defaults to 0.3.
        seed : Random generator seed. Defaults to 42.
        num_workers: Number of workers for dataloaders. Defaults to 6.
        train_transform: Transformations for train dataset.
            Defaults to None.
        val_transform : Transformations for validation dataset.
            Defaults to None.
        test_transform : Transformations for test dataset.
            Defaults to None.
        train_split_file: path to txt file with training samples list
        val_split_file: path to txt file with validation samples list
        test_split_file: path to txt file with test samples list
        exclude_good : If True, exclude good samples from the dataset. Defaults to False.
        num_data_train: number of samples to use in the train split (shuffle the samples and pick the
            first num_data_train)
        one_hot_encoding: if True, the labels are one-hot encoded to N channels, where N is the number of classes.
            If False, masks are single channel that contains values as class indexes. Defaults to True.
    """

    def __init__(
        self,
        data_path: str,
        idx_to_class: dict,
        name: str = "multiclass_segmentation_datamodule",
        dataset: type[SegmentationDatasetMulticlass] = SegmentationDatasetMulticlass,
        batch_size: int = 32,
        test_size: float = 0.3,
        val_size: float = 0.3,
        seed: int = 42,
        num_workers: int = 6,
        train_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        train_split_file: str | None = None,
        test_split_file: str | None = None,
        val_split_file: str | None = None,
        exclude_good: bool = False,
        num_data_train: int | None = None,
        one_hot_encoding: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            batch_size=batch_size,
            num_workers=num_workers,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            **kwargs,
        )
        self.test_size = test_size
        self.val_size = val_size
        self.exclude_good = exclude_good
        self.train_split_file = train_split_file
        self.test_split_file = test_split_file
        self.val_split_file = val_split_file
        self.dataset = dataset
        self.idx_to_class = idx_to_class
        self.num_data_train = num_data_train
        self.one_hot_encoding = one_hot_encoding
        self.train_dataset: SegmentationDataset
        self.val_dataset: SegmentationDataset
        self.test_dataset: SegmentationDataset

    def _preprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        """Function to preprocess the mask.

        Args:
            mask: a numpy array of dimension HxW with values in [0] + self.idx_to_class.

        Output:
            a binary numpy array with dims len(self.idx_to_class+1)xHxW
        """
        # For each class we must have a channel
        multilayer_mask = np.zeros((len(self.idx_to_class) + 1, *mask.shape[:2]))
        for idx in self.idx_to_class:
            multilayer_mask[int(idx)] = (mask == int(idx)).astype(np.uint8)

        return multilayer_mask

    def _resolve_label(self, path: str) -> np.ndarray:
        """Return a binary array of 1 + len(self.idx_to_class) with 1 if that class is present in the mask."""
        one_hot = np.zeros([len(self.idx_to_class) + 1], np.uint8)  # add class 0
        mask = cv2.imread(path, 0)
        if mask.sum() == 0:
            one_hot[0] = 1
        else:
            indices = np.unique(mask)
            one_hot[indices] = 1
            one_hot[0] = 0

        return one_hot

    def _read_folder(self, data_path: str) -> tuple[list[str], list[np.ndarray], list[str]]:
        """Read a folder containing images and masks subfolders.

        Args:
            data_path: Path to the data folder.

        Returns:
            List of paths to the images, list of associated one-hot encoded targets and list of mask paths.
        """
        samples = []
        targets = []
        masks = []

        for im in glob.glob(os.path.join(data_path, "images", "*")):
            if im[0] == ".":
                continue

            mask_path = glob.glob(os.path.splitext(im.replace("images", "masks"))[0] + ".*")

            if len(mask_path) == 0:
                log.debug("Mask not found: %s", os.path.basename(im))
                continue

            if len(mask_path) > 1:
                raise ValueError(f"Multiple masks found for image: {os.path.basename(im)}, this is not supported")

            target = self._resolve_label(mask_path[0])
            samples.append(im)
            targets.append(target)
            masks.append(mask_path[0])

        return samples, targets, masks

    def _read_split(self, split_file: str) -> tuple[list[str], list[np.ndarray], list[str]]:
        """Reads split file.

        Args:
            split_file: Path to the split file.

        Returns:
            List of paths to images, labels and mask paths.
        """
        samples, targets, masks = [], [], []
        with open(split_file) as f:
            split = f.read().splitlines()
        for sample in split:
            sample_path = os.path.join(self.data_path, sample)
            mask_path = glob.glob(os.path.splitext(sample_path.replace("images", "masks"))[0] + ".*")

            if len(mask_path) == 0:
                log.debug("Mask not found: %s", os.path.basename(sample_path))
                continue

            if len(mask_path) > 1:
                raise ValueError(
                    f"Multiple masks found for image: {os.path.basename(sample_path)}, this is not supported"
                )

            target = self._resolve_label(mask_path[0])
            samples.append(sample_path)
            targets.append(target)
            masks.append(mask_path[0])

        return samples, targets, masks

    def _prepare_data(self) -> None:
        """Prepare data for training and testing."""
        if not (self.train_split_file and self.test_split_file and self.val_split_file):
            all_samples, all_targets, all_masks = self._read_folder(self.data_path)

            (
                samples_and_masks_train,
                targets_train,
                samples_and_masks_test,
                targets_test,
            ) = iterative_train_test_split(
                np.expand_dims(np.array(list(zip(all_samples, all_masks, strict=False))), 1),
                np.array(all_targets),
                test_size=self.test_size,
            )

            samples_train, samples_test = samples_and_masks_train[:, 0, 0], samples_and_masks_test[:, 0, 0]
            masks_train, masks_test = samples_and_masks_train[:, 0, 1], samples_and_masks_test[:, 0, 1]

        if self.test_split_file:
            samples_test, targets_test, masks_test = self._read_split(self.test_split_file)
            if not self.train_split_file:
                samples_train, targets_train, masks_train = [], [], []
                for sample, target, mask in zip(all_samples, all_targets, all_masks, strict=False):
                    if sample not in samples_test:
                        samples_train.append(sample)
                        targets_train.append(target)
                        masks_train.append(mask)

        if self.train_split_file:
            samples_train, targets_train, masks_train = self._read_split(self.train_split_file)
            if not self.test_split_file:
                samples_test, targets_test, masks_test = [], [], []
                for sample, target, mask in zip(all_samples, all_targets, all_masks, strict=False):
                    if sample not in samples_train:
                        samples_test.append(sample)
                        targets_test.append(target)
                        masks_test.append(mask)

        if self.val_split_file:
            samples_val, targets_val, masks_val = self._read_split(self.val_split_file)
            if not self.test_split_file or not self.train_split_file:
                raise ValueError("Validation split file is specified but no train or test split file is specified.")
        else:
            samples_and_masks_train, targets_train, samples_and_masks_val, targets_val = iterative_train_test_split(
                np.expand_dims(np.array(list(zip(samples_train, masks_train, strict=False))), 1),
                np.array(targets_train),
                test_size=self.val_size,
            )
            samples_train = samples_and_masks_train[:, 0, 0]
            samples_val = samples_and_masks_val[:, 0, 0]
            masks_train = samples_and_masks_train[:, 0, 1]
            masks_val = samples_and_masks_val[:, 0, 1]

        # Pre-ordering train and val samples for determinism
        # They will be shuffled (with a seed) during training
        sorting_indices_train = np.argsort(list(samples_train))
        samples_train = [samples_train[i] for i in sorting_indices_train]
        targets_train = [targets_train[i] for i in sorting_indices_train]
        masks_train = [masks_train[i] for i in sorting_indices_train]

        sorting_indices_val = np.argsort(samples_val)
        samples_val = [samples_val[i] for i in sorting_indices_val]
        targets_val = [targets_val[i] for i in sorting_indices_val]
        masks_val = [masks_val[i] for i in sorting_indices_val]

        if self.exclude_good:
            samples_train = list(np.array(samples_train)[np.array(targets_train)[:, 0] == 0])
            masks_train = list(np.array(masks_train)[np.array(targets_train)[:, 0] == 0])
            targets_train = list(np.array(targets_train)[np.array(targets_train)[:, 0] == 0])

        if self.num_data_train is not None:
            # Generate a random permutation
            random_permutation = list(range(len(samples_train)))
            random.seed(self.seed)
            random.shuffle(random_permutation)

            # Shuffle samples_train, targets_train, and masks_train using the same permutation
            samples_train = [samples_train[i] for i in random_permutation]
            targets_train = [targets_train[i] for i in random_permutation]
            masks_train = [masks_train[i] for i in random_permutation]

            samples_train = np.array(samples_train)[: self.num_data_train]
            targets_train = np.array(targets_train)[: self.num_data_train]
            masks_train = np.array(masks_train)[: self.num_data_train]

        df_list = []
        for split_name, samples, targets, masks in [
            ("train", samples_train, targets_train, masks_train),
            ("val", samples_val, targets_val, masks_val),
            ("test", samples_test, targets_test, masks_test),
        ]:
            df = pd.DataFrame({"samples": samples, "targets": list(targets), "masks": masks})
            df["split"] = split_name
            df_list.append(df)

        self.data = pd.concat(df_list, axis=0)

    def setup(self, stage=None):
        """Setup data module based on stages of training."""
        if stage in ["fit", "train"]:
            train_data = self.data[self.data["split"] == "train"]
            val_data = self.data[self.data["split"] == "val"]

            self.train_dataset = self.dataset(
                image_paths=train_data["samples"].tolist(),
                mask_paths=train_data["masks"].tolist(),
                idx_to_class=self.idx_to_class,
                transform=self.train_transform,
                one_hot=self.one_hot_encoding,
            )
            self.val_dataset = self.dataset(
                image_paths=val_data["samples"].tolist(),
                mask_paths=val_data["masks"].tolist(),
                transform=self.val_transform,
                idx_to_class=self.idx_to_class,
                one_hot=self.one_hot_encoding,
            )
        elif stage == "test":
            self.test_dataset = self.dataset(
                image_paths=self.data[self.data["split"] == "test"]["samples"].tolist(),
                mask_paths=self.data[self.data["split"] == "test"]["masks"].tolist(),
                transform=self.test_transform,
                idx_to_class=self.idx_to_class,
                one_hot=self.one_hot_encoding,
            )
        elif stage == "predict":
            pass
        else:
            raise ValueError(f"Unknown stage {stage}")

    def train_dataloader(self) -> DataLoader:
        """Returns the train dataloader.

        Raises:
            ValueError: If train dataset is not initialized.

        Returns:
            Train dataloader.
        """
        if not self.train_dataset_available:
            raise ValueError("Train dataset is not initialized")

        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def val_dataloader(self) -> DataLoader:
        """Returns the validation dataloader.

        Raises:
            ValueError: If validation dataset is not initialized.

        Returns:
            val dataloader.
        """
        if not self.val_dataset_available:
            raise ValueError("Validation dataset is not initialized")

        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def test_dataloader(self) -> DataLoader:
        """Returns the test dataloader.

        Raises:
            ValueError: If test dataset is not initialized.


        Returns:
            test dataloader.
        """
        if not self.test_dataset_available:
            raise ValueError("Test dataset is not initialized")

        loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return self.test_dataloader()



import torch
import torch.nn.functional as F


def simsiam_loss(
    p1: torch.Tensor,
    p2: torch.Tensor,
    z1: torch.Tensor,
    z2: torch.Tensor,
) -> torch.Tensor:
    """SimSIAM loss described in https://arxiv.org/abs/2011.10566.

    Args:
        p1: First `predicted` features (i.e. h(f(T(x1))))
        p2: Second `predicted` features (i.e. h(f(T'(x2))))
        z1: First 'projected features (i.e. f(T(x1)))
        z2: Second 'projected features (i.e. f(T(x2)))

    Returns:
        SimSIAM loss
    """
    return -(F.cosine_similarity(p1, z2).mean() + F.cosine_similarity(p2, z1).mean()) * 0.5


class SimSIAMLoss(torch.nn.Module):
    """SimSIAM loss module."""

    def forward(self, p1: torch.Tensor, p2: torch.Tensor, z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Compute the SimSIAM loss."""
        return simsiam_loss(p1, p2, z1, z2)



from __future__ import annotations

import os
from typing import Any

import albumentations
import cv2
import numpy as np
import pandas as pd
from torchvision.datasets.utils import download_and_extract_archive

from quadra.datamodules import SegmentationMulticlassDataModule
from quadra.datasets.segmentation import SegmentationDatasetMulticlass
from quadra.utils import utils

log = utils.get_logger(__name__)


class OxfordPetSegmentationDataModule(SegmentationMulticlassDataModule):
    """OxfordPetSegmentationDataModule.

    Args:
        data_path: path to the oxford pet dataset
        idx_to_class: dict with corrispondence btw mask index and classes: {1: class_1, 2: class_2, ..., N: class_N}
            except background class which is 0.
        name: Defaults to "oxford_pet_segmentation_datamodule".
        dataset: Defaults to SegmentationDataset.
        batch_size:  batch size for training. Defaults to 32.
        test_size: Defaults to 0.3.
        val_size:  Defaults to 0.3.
        seed: Defaults to 42.
        num_workers: number of workers for data loading. Defaults to 6.
        train_transform: Train transform. Defaults to None.
        test_transform: Test transform. Defaults to None.
        val_transform: Validation transform. Defaults to None.
    """

    def __init__(
        self,
        data_path: str,
        idx_to_class: dict,
        name: str = "oxford_pet_segmentation_datamodule",
        dataset: type[SegmentationDatasetMulticlass] = SegmentationDatasetMulticlass,
        batch_size: int = 32,
        test_size: float = 0.3,
        val_size: float = 0.3,
        seed: int = 42,
        num_workers: int = 6,
        train_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            idx_to_class=idx_to_class,
            name=name,
            dataset=dataset,
            batch_size=batch_size,
            test_size=test_size,
            val_size=val_size,
            seed=seed,
            num_workers=num_workers,
            train_transform=train_transform,
            test_transform=test_transform,
            val_transform=val_transform,
            **kwargs,
        )

    _RESOURCES = (
        ("https://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz", "5c4f3ee8e5d25df40f4fd59a7f44e54c"),
        ("https://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz", "95a8c909bbe2e81eed6a22bccdf3f68f"),
    )

    def _preprocess_mask(self, mask: np.ndarray) -> np.ndarray:
        """Preprocess mask function that is adapted from
        https://albumentations.ai/docs/examples/pytorch_semantic_segmentation/.

        Args:
            mask: mask to be preprocessed

        Returns:
            binarized mask
        """
        mask = mask.astype(np.float32)
        mask[mask == 2.0] = 0.0
        mask[(mask == 1.0) | (mask == 3.0)] = 1.0
        mask = (mask > 0).astype(np.uint8)
        return mask

    def _check_exists(self, image_folder: str, annotation_folder: str) -> bool:
        """Check if the dataset is already downloaded."""
        return all(os.path.exists(folder) and os.path.isdir(folder) for folder in (image_folder, annotation_folder))

    def download_data(self):
        """Download the dataset if it is not already downloaded."""
        image_folder = os.path.join(self.data_path, "images")
        annotation_folder = os.path.join(self.data_path, "annotations")
        if not self._check_exists(image_folder, annotation_folder):
            for url, md5 in self._RESOURCES:
                download_and_extract_archive(url, download_root=self.data_path, md5=md5, remove_finished=True)
            log.info("Fixing corrupted files...")
            images_filenames = sorted(os.listdir(image_folder))
            for filename in images_filenames:
                file_wo_ext = os.path.splitext(os.path.basename(filename))[0]
                try:
                    mask = cv2.imread(os.path.join(annotation_folder, "trimaps", file_wo_ext + ".png"))
                    mask = self._preprocess_mask(mask)
                    if np.sum(mask) == 0:
                        os.remove(os.path.join(image_folder, filename))
                        os.remove(os.path.join(annotation_folder, "trimaps", file_wo_ext + ".png"))
                        log.info("Removed %s", filename)
                    else:
                        img = cv2.imread(os.path.join(image_folder, filename))
                        cv2.imwrite(os.path.join(image_folder, file_wo_ext + ".jpg"), img)
                except Exception:
                    ip = os.path.join(image_folder, filename)
                    mp = os.path.join(annotation_folder, "trimaps", file_wo_ext + ".png")
                    if os.path.exists(ip):
                        os.remove(ip)
                    if os.path.exists(mp):
                        os.remove(mp)
                    log.info("Removed %s", filename)

    def _prepare_data(self) -> None:
        """Prepare the data to be used by the DataModule."""
        self.download_data()

        trainval_split_filepath = os.path.join(self.data_path, "annotations", "trainval.txt")
        with open(trainval_split_filepath) as f:
            split_data = f.read().strip("\n").split("\n")
        trainval_filenames = [
            x.split(" ")[0]
            for x in split_data
            if os.path.exists(os.path.join(self.data_path, "images", x.split(" ")[0] + ".jpg"))
        ]
        train_filenames = [x for i, x in enumerate(trainval_filenames) if i % 10 != 0]
        val_filenames = [x for i, x in enumerate(trainval_filenames) if i % 10 == 0]

        test_split_filepath = os.path.join(self.data_path, "annotations", "test.txt")
        with open(test_split_filepath) as f:
            split_data = f.read().strip("\n").split("\n")
        test_filenames = [
            x.split(" ")[0]
            for x in split_data
            if os.path.exists(os.path.join(self.data_path, "images", x.split(" ")[0] + ".jpg"))
        ]

        df_list = []
        for split_name, filenames in [
            ("train", train_filenames),
            ("val", val_filenames),
            ("test", test_filenames),
        ]:
            samples = [os.path.join(self.data_path, "images", f + ".jpg") for f in filenames]
            masks = [os.path.join(self.data_path, "annotations", "trimaps", f + ".png") for f in filenames]
            targets = [1] * len(filenames)

            df = pd.DataFrame({"samples": samples, "masks": masks, "targets": targets})
            df["split"] = split_name
            df_list.append(df)

        self.data = pd.concat(df_list, axis=0)



import numpy as np
import skimage
from skimage.morphology import medial_axis

from quadra.utils import utils

log = utils.get_logger(__name__)


def smooth_mask(mask: np.ndarray) -> np.ndarray:
    """Smooths for segmentation.

    Args:
        mask: Input mask

    Returns:
        Smoothed mask
    """
    labeled_mask = skimage.measure.label(mask)
    labels = np.arange(0, np.max(labeled_mask) + 1)
    output_mask = np.zeros_like(mask).astype(np.float32)
    for l in labels:
        component_mask = labeled_mask == l
        _, distance = medial_axis(component_mask, return_distance=True)
        component_mask_norm = distance ** (1 / 2.2)
        component_mask_norm = (component_mask_norm - np.min(component_mask_norm)) / (
            np.max(component_mask_norm) - np.min(component_mask_norm)
        )
        output_mask += component_mask_norm
    output_mask = output_mask * mask
    return output_mask



from __future__ import annotations

from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler


class LearningRateScheduler(_LRScheduler):
    """Provides inteface of learning rate scheduler.

    Note:
        Do not use this class directly, use one of the sub classes.
    """

    def __init__(self, optimizer: Optimizer, init_lr: tuple[float, ...]):
        # pylint: disable=super-init-not-called
        self.optimizer = optimizer
        self.init_lr = init_lr

    def step(self, *args, **kwargs):
        """Base method, must be implemented by the sub classes."""
        raise NotImplementedError

    def set_lr(self, lr: tuple[float, ...]):
        """Set the learning rate for the optimizer."""
        if self.optimizer is not None:
            for i, g in enumerate(self.optimizer.param_groups):
                if "fix_lr" in g and g["fix_lr"]:
                    if len(lr) == 1:
                        lr_to_set = self.init_lr[0]
                    else:
                        lr_to_set = self.init_lr[i]
                elif len(lr) == 1:
                    lr_to_set = lr[0]
                else:
                    lr_to_set = lr[i]
                g["lr"] = lr_to_set

    def get_lr(self):
        """Get the current learning rate if the optimizer is available."""
        if self.optimizer is not None:
            for g in self.optimizer.param_groups:
                return g["lr"]

        return None



from .barlowtwins import BarlowTwins
from .byol import BYOL
from .dino import Dino
from .idmm import IDMM
from .simclr import SimCLR
from .simsiam import SimSIAM
from .vicreg import VICReg

__all__ = [
    "BarlowTwins",
    "BYOL",
    "Dino",
    "IDMM",
    "SimCLR",
    "SimSIAM",
    "VICReg",
]
